*  FUNCTION gMatrix for FoxPro
*  SYNTAX:
*  IF gMatrix( 'table.field_name' )
*    *--whatever you need to do...
*  NOTE: you must pass the table.fieldname as a literal string as above, 
*  otherwise it won't work and will crash!
*  generates binary matrix table for to be used in a SQL statement that involves 
*  percent, ratio & longitud. difference calculations and other hairy problems
*  that might ordinarily require sequential processing or multiple SQLs or subselects
*
*  Author: Lee Rodgers, Austin, Tx. 73144.2276@CompuServe.COM / leebert@io.com
*
*  SEE EXAMPLES AT BOTTOM OF FILE
*  examples included paraphrase DBMS Nov. 1994 SQL/Server article by 
*  Steve Roti (CompuServe:70323.3614@CompuServe.COM)
*
PARAMETERS s_fld

PRIVATE _s_alias, a_columns, a_rows, _l_AutoVals, _sql_statement, _sAppendFrom

IF TYPE(s_fld) = "U" AND !(OCCURS( '.', s_fld ) > 0 )
    ?? CHR(7)
    ?? CHR(7)
    WAIT WINDOW "ERROR - THE APPROPRIATE TABLE IS NOT SELECTED. NO MATRIX TABLE GENERATED"
    RETURN .F.
ENDIF

_s_alias = ALIAS()
IF OCCURS( '.', s_fld ) > 0 
  _s_alias = Parser( s_fld, 1, '.' )
  IF !db_open( _s_alias )
    ?? CHR(7)
    ?? CHR(7)
    WAIT WINDOW "ERROR - THE APPROPRIATE TABLE COULD NOT BE OPENED. NO MATRIX TABLE GENERATED"
    RETURN .F.
  ENDIF
ENDIF

IF EMPTY( _s_alias )
  ?? CHR(7)
  ?? CHR(7)
  WAIT WINDOW "ERROR - THE APPROPRIATE TABLE COULD NOT BE OPENED. NO MATRIX TABLE GENERATED"
  RETURN .F.
ELSE
  SELECT (_s_alias )
ENDIF  

s_aliased_fld = IIF( OCCURS( '.', s_fld )>0, s_fld, ALIAS() + '.' + s_fld )
s_fld = IIF( OCCURS( '.', s_fld )>0, Parser( s_fld, 2, '.' ), s_fld ) 

_l_num_key = .F.
IF ISDIGIT( EVALUATE( s_aliased_fld ) )
  IF LEN(EVALUATE( s_aliased_Fld )) <= 9
    *--add alpha letter like 'Z' to beginning of field
    _l_num_key = .T.
  ELSE
    ?? CHR(7)
    ?? CHR(7)
    WAIT WINDOW "ERROR - PLEASE RESPECIFY YOUR KEY SO THAT THE FIRST CHARACTER OF THE FIELD" + CHR(13) + "IS ALPHA. PURE NUMERIC NAMES ARE NOT ALLOWED BECAUSE OF FOXPRO. NO MATRIX TABLE GENERATED"
    RETURN .F.
  ENDIF
ENDIF  

IF LEN(EVALUATE(s_aliased_fld)) > 10
  ?? CHR(7)
  ?? CHR(7)
  WAIT WINDOW "ERROR - YOUR KEY EXCEEDS 10 CHARACTERS IN LENGTH: INFORMATION WOULD BE LOST BECAUSE OF FOXPRO" + CHR(13) + "NO MATRIX TABLE GENERATED"
  RETURN .F.
ENDIF  

_sql_statement = ; 
[SELECT DISTINCT(] + IIF(_l_num_key,["Z"+],[]) ;
+ s_aliased_fld + [) AS field_name, "N" AS field_type, 1 AS field_len, 0 AS field_dec ] ;
+ [ FROM ] + ALIAS() + ;
[ GROUP BY 1 ORDER BY 1 INTO CURSOR get_rows ]

&_sql_statement

IF RECCOUNT( 'get_rows' ) > 254
  ?? CHR(7)
  ?? CHR(7)
  WAIT WINDOW NOWAIT "ERROR - TOO MANY DISTINCT ROWS - EXCEEDS 254." + CHR(13) + "CANNOT CREATE MORE THAN 254 COLUMNS FOR MATRIX TABLE" + CHR(13) + "NO MATRIX TABLE GENERATED"
  USE IN get_rows
  RETURN .F.
ENDIF

IF RECCOUNT( 'get_rows' ) = 0
  ?? CHR(7)
  ?? CHR(7)
  WAIT WINDOW NOWAIT "ERROR - ZERO ROWS" + CHR(13) + "CANNOT MATRIX NOTHING" + CHR(13) + "NO MATRIX TABLE GENERATED"
  USE IN get_rows
  RETURN .F.
ENDIF

CREATE CURSOR get_column (field_name C(10), field_type C(1), field_len N(3), field_dec N(3) )
SELECT get_column
INSERT INTO get_column ( field_name, field_type, field_len, field_dec) ;
                 VALUES ( s_fld, TYPE(s_aliased_fld), LEN(EVALUATE(s_aliased_fld))+IIF(_l_num_key,1,0), 0 )
_sAppendFrom = DBF('get_rows')
APPEND FROM ( _sAppendFrom )

SELECT 0
_sTempDbf = LEFT( "Z" + SYS(3), 8 )
CREATE ( _sTempDbf ) FROM (DBF( 'get_column' ))
IF USED('matrix')
 USE IN matrix
ENDIF
SELECT (_sTempDbf )
USE ( _sTempDbf ) ALIAS matrix 

SELECT field_name FROM get_rows INTO ARRAY a_rows ORDER BY 1 
INSERT INTO matrix FROM ARRAY a_rows

SELECT Matrix

SCAN
  s_curr_fld = "matrix."+ EVALUATE( s_fld )
  REPLACE &s_curr_fld WITH 1
ENDSCAN

RETURN .T.

*--------------------------------------------------------------
*
* FUNCTION parser
* Parses strings based upon current value of _n_posit from a 
* string that looks like this "AB,CD,EF,GH" 
* returning AB when passed 1, CD when passed 2, etc...
*--------------------------------------------------------------
FUNCTION parser
PARAMETERS _s_string, _n_posit, _s_delimiter
PRIVATE n_beg_pos, n_end_pos, _s_working_str

n_beg_pos = 1
n_end_pos = 0
_s_working_str = ''

IF INLIST( TYPE( '_s_delimiter' ), "U","L" )
  _s_delimiter = ','
ENDIF

*parse out table names from _s_string string

_n_segments = OCCURS( _s_delimiter , _s_string ) + 1
IF _n_Posit > _n_segments
*  *--they've specified an invalid position on the string
*  ?? CHR(7)
*  ?? CHR(7)
*  WAIT WINDOW "THE SEGMENT YOU SPECIFIED TO PARSE DOES NOT EXIST!" + CHR(13) + "YOU SPECIFIED BEYOND THE RANGE OF SEGMENTS!" + CHR(13) + "PARSER() WILL RETURN A NULL STRING!"
  RETURN ''  && return null string
ENDIF


n_beg_pos = AT( _s_delimiter  , _s_string, IIF(_n_posit>1,_n_posit-1,1) )
IF _n_posit = 0
  *--user passed 0!
  n_end_pos = 0
ELSE  
  n_end_pos = AT( _s_delimiter , _s_string, _n_posit )
ENDIF  

  DO CASE
    CASE n_end_pos = 0 AND _n_posit > 1
    *--go to end of string
      _s_parsed_string = SUBSTR( _s_string, n_beg_pos+1 )
    CASE n_end_pos = 0 AND _n_posit = 1 
      _s_parsed_string = STRTRAN(_s_string,_s_delimiter,"")
    CASE n_end_pos > 1 AND _n_posit = 1
      _s_parsed_string = SUBSTR( _s_string, 1, (n_end_pos-1) )
    CASE n_end_pos > 1 AND _n_posit > 1
      _s_parsed_string = SUBSTR( _s_string, n_beg_pos+1, (n_end_pos-n_beg_pos)-1 )
    CASE _n_posit = 0
      WAIT WINDOW NOWAIT "ZERO PASSED AS 2nd PARAMETER"
      _s_parsed_string = ''
  ENDCASE

RETURN _s_parsed_string

**-----------------------------------------------------------------
*
*               FUNCTION db_open
*
*      SYNTAX:
*      IF db_open ( <expC1> ; database name
*                   <expC2> ; index tag
*                   <expN1> ; 0 - 225 work area
*                   <expC3> ; alias
*                   <expN2> ; timeout on file use attempts in seconds
*                   <expL1> ; by reference, did this table get opened? (for chained F() calls to db_open())
*                   <expL2> ; open it noupdate (.T.) or normal (.F.)
*
*  calls: Net_use()
*
*
**-----------------------------------------------------------------
FUNCTION db_open 
PARAMETERS s_db_name, s_order, n_workarea,  ;
	   s_alias, n_timeout, l_WasOpened, l_no_update

IF TYPE( 'n_timeout' ) # "N"
  n_timeout = 0
ENDIF
IF TYPE( 'n_workarea') # "N"
  n_workarea = 0
ENDIF  
IF TYPE( 's_alias' ) # "C" OR EMPTY( s_alias )
  s_alias = STRTRAN( s_db_name, ".DBF", "" )
  s_alias = IIF( "\" $ s_db_name, RIGHT( s_db_name, RAT( "\", ALLTRIM(s_db_name))-2), s_db_name )
ENDIF

IF TYPE( 's_order' ) # "C"
  s_order  = ''
ENDIF  
n_start_time = VAL( DTOS( DATE() ) - SYS(2) )


IF !USED( s_alias ) 
  *SELECT ( n_workarea )  && don't switch work areas, slows down system
  l_ret_val = net_use( s_db_name, l_no_update , n_timeout, s_alias, n_workarea )
ELSE
  *SELECT ( s_alias )  && don't switch work areas, slows down system
  l_ret_val = .T.
ENDIF  
     

IF !EMPTY( s_order ) 
  SET ORDER TO ( s_order ) ASCENDING IN (s_alias)
ENDIF   

RETURN l_ret_val
*-------------------------------------------------------------------
*
*
*      FUNCTION NET_USE
*  SYNTAX:  <EXPC1> ; dbf file
*           <EXPL1> ; noupdate or not (default is .F. )
*           <EXPN1> ; seconds to retry
*           <EXPC2> ; alias (optional)
*            <EXPN2> ; workarea
*
*
*
*-------------------------------------------------------------------
FUNCTION NET_USE
PARAMETERS s_file, l_no_update, n_wait, s_alias, n_workarea

IF TYPE( 's_alias' ) # "C"
  s_alias = STRTRAN( s_file, ".DBF", "" )
  s_alias = IIF( "\" $ s_alias, ;
		      RIGHT( s_alias , RAT( "\", s_alias )-2 ), ;
		      s_alias ;
	       )
ENDIF

s_on_error = ON("ERROR")
n_start_time = VAL( DTOS( DATE() ) - SYS(2) )
l_error_exists = .F.
ON ERROR DO netuse_err WITH ERROR(), MESSAGE(), MESSAGE(1), PROGRAM(), LINENO()

DO WHILE .T.
   
 l_error_exists = .F.

 IF l_no_update
    USE (LOCFILE(s_file,"DBF","Where is "+s_file+"?")) AGAIN NOUPDATE SHARED ALIAS (s_alias) IN (n_workarea) 
 ELSE
    USE (LOCFILE( s_file, "DBF", [Where is ] + s_file + [?] )) AGAIN SHARED ALIAS (s_alias) IN (n_workarea)
 ENDIF
 
 IF !l_error_exists
   l_ret_val = .T.
   EXIT
 ELSE 
   IF n_wait > 0 AND (VAL( DTOS( DATE() ) - SYS(2) )) - (n_start_time) > n_wait
     l_ret_val = .F.
     EXIT
   ENDIF
 ENDIF
 *-- wait 1 second
 WAIT TIMEOUT n_timeout WINDOW 'RE-ATTEMPTING FILE USE'
 IF LASTKEY() = 27
   l_ret_val = .F.
   EXIT
 ENDIF                         
ENDDO

ON ERROR &s_on_error
RETURN  l_ret_val
*-----------------------------------------------------------------
*
*
*
*      PROC netuse_err
*
*
*
*-----------------------------------------------------------------
PROC netuse_err
PARAMETERS _error, _message, _message1, _program, _lineno  
l_error_exists = .T.
IF _error = 1113
  ?? CHR(7)
  ?? CHR(7)
  WAIT WINDOW 'rebuild indices, baby!'
  n_wait = 1
ELSE
  WAIT WINDOW _message + _Message1
*  SET STEP ON
ENDIF
RETURN



*=====================================================================================
*=====================================================================================
*=====================================================================================
*=====================================================================================
*=====================================================================================
* EXAMPLES
*=====================================================================================
*=====================================================================================
*=====================================================================================
*=====================================================================================
*Datamation article complains SQL can't do anything requiring 'sequential processing,' 
*including *percents, ratios, or differences. However, with binary matrix, SQL can handle 
*these, described *as 'item dimension or time dimension' queries. To handle these, you must 
*use a binary matrix.
*
*Examples:

=creatable1() && populate fact & matrix tables

*1. Sales ratio of Alice's products vs. Julies' products
SELECT SUM(Alice*Dollars) AS Alices, ;
       SUM(Julie*Dollars) AS Julies, ;
       SUM(alice*dollars)/SUM(julie*dollars) AS SalesRatio ;
       FROM fact, matrix ;
       WHERE fact.person == matrix.person ;
       INTO CURSOR result1
SELECT result1       
Brow nowait

*2. Sales percentage of Alice vs. total
SELECT SUM(Alice*Dollars) AS Alices, ;
   SUM(Dollars) AS Total_Prod, ;
   SUM(Alice*Dollars)/Sum(Dollars)*100 ;
   AS Sales_Percnt ;
   FROM fact, matrix ;
   WHERE fact.person == matrix.person ;
   INTO CURSOR result2
SELE result2
Brow nowait

*3. Promo'd vs. non-Promo'd products
*--build a promotionals matrix
CREATE CURSOR Promo (Promo C(1), Yes N(1), No N(1) )
APPEND BLANK 
REPLACE Promo WITH "Y", Yes WITH 1, No WITH 0
APPEND BLANK 
REPLACE Promo WITH "N", Yes WITH 0, No WITH 1
*run query
SELECT SUM(yes*Dollars) AS Promo, ;
       SUM(no*Dollars) AS Non_Promo, ;
       SUM(Yes*Dollars)/SUM(No*Dollars) AS ratio ;
       FROM Fact, Promo ;
       WHERE Fact.promo == Promo.promo ;
       INTO CURSOR Promo_out
Brow nowait

*--4. analyze date ranges
*--build a year matrix
CREATE CURSOR Years (Year n(4), This_year N(1), Last_year N(1) )
APPEND BLANK 
REPLACE year WITH 1994, This_year WITH 1, Last_year WITH 0
APPEND BLANK 
REPLACE year WITH 1993, This_year WITH 0, Last_year WITH 1

*--looking @ xmas sales, etc.
SELECT SUM(this_year*dollars) AS this_year, ;
       SUM(last_year*dollars) AS last_year ;
       FROM fact, years ;
       WHERE YEAR(saledate) = years.year ;
       AND MONTH(saledate) = 12 ;
       INTO CURSOR year_out1
Brow nowait

*--looking at YTD(year to date) sales
SELECT SUM(this_year*dollars) AS this_YTD, ;
       SUM(last_year*dollars) AS last_YTD ;
       FROM fact, years ;
       WHERE YEAR(saledate) = years.year ;
       AND ( ;
       (YEAR(saledate)=YEAR(DATE()) AND saledate<=DATE() ) ;
       OR ;
       (YEAR(fact.saledate) = YEAR(DATE()-365) AND saledate<=(DATE()-365)) ;
       ) ;
       INTO TABLE Year_out2
Brow nowait       
       
*=================================================================
FUNCTION creatable1
CREATE CURSOR fact (person C(5), customer C(5), product C(6), Saledate D, dollars N(7,2), units N(2), Promo C(1) )
DECLARE a_temp[ 4, 7 ]
a_t